/***************************************************************************
                          pattern.cpp  -  description
                             -------------------
    begin                : Wed Jun 26 2002
    copyright            : (C) 2002 by Nobuhiro Kuroiwa
    email                : nobu@iastate.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <list>
#include <iostream>

#include "SDL/SDL.h"
#include "SDL/SDL_mixer.h"

#include "pattern.h"
#include "stranlz.h"
#include "crand.h"


list<Alies*>			lst_table_alies;
list<Alies*>::iterator	it_table_alies;
extern bool				stopAllThread;

//---------------------------------------------------------------------------
//	Default Constructor of Pattern Class
//---------------------------------------------------------------------------
Pattern::Pattern():
ppattern(NULL), own_ppat(true)
{
}

//---------------------------------------------------------------------------
//	Constructor of Pattern Class
//---------------------------------------------------------------------------
Pattern::Pattern( string *pstr_newname)
{
	name = *pstr_newname;
}

//---------------------------------------------------------------------------
//	Member Function of Pattern Class
//---------------------------------------------------------------------------
void Pattern::play()
{
	if( ppattern != NULL && !stopAllThread){	ppattern->play();}
}
//---------------------------------------------------------------------------
//	Member Function of Pattern Class
//---------------------------------------------------------------------------
void Pattern::setPosition4Note( Sint16 new_angle, Uint8 new_dist)
{
	if( ppattern != NULL)	ppattern->setPosition4Note( new_angle, new_dist);
}

//---------------------------------------------------------------------------
//	Member Function of Pattern Class
//---------------------------------------------------------------------------
Pattern::~Pattern(){
	if( ppattern != NULL && own_ppat == true)
	{
		ppattern->~Pattern();
	}
}

//---------------------------------------------------------------------------
//	Member Function of Pattern Class
//---------------------------------------------------------------------------
bool Pattern::growTree()
{
	pealParenthesis( &name);

	switch( diagnoseType( &name))


	{
	case	TYPE_DIAGNOSE_PATTERN	://	cout << name << " is Pattern" << endl;
				ppattern = new Pattern( &name);
				break;
	case	TYPE_DIAGNOSE_NOTE		:	//cout << name << " is Note" << endl;
				ppattern = new Note( &name);
				break;
	case	TYPE_DIAGNOSE_RANDOMREST:	//cout << name << " is RandRest" << endl;
				ppattern = new RandomRest( &name);
				break;
	case	TYPE_DIAGNOSE_MULTIPATTERN:	//cout << name << " is MP" << endl;
				ppattern = new MultiPattern( &name);
				break;
	case	TYPE_DIAGNOSE_PROBABILITYPATTERN:
                                      	//cout << name << " is PP" << endl;
				ppattern = new ProbabilityPattern( &name);
				break;
	case	TYPE_DIAGNOSE_REST		:	//cout << name << " is Rest" << endl;
				ppattern = new Rest( &name);
				break;
	case TYPE_DIAGNOSE_LABEL		:
				// original one is stored in lst_table_psong
				// so this ppattern shouldn't deleted in this
				// class.
				ppattern = searchSongList( &name);

				if( ppattern != NULL)
				{						//cout << name << " is label" << endl;
					own_ppat = false;
					return true;
				}
				// No break, go to default:
	default							:
				cout << "can't understand as Pattern Object \""
				+ name + "\"" <<endl;
				return false;
	}
	own_ppat = true;
	return ppattern->growTree();	
}

//---------------------------------------------------------------------------
//	Member Function of Pattern Class
//---------------------------------------------------------------------------
string Pattern::getName() const

{
	return name;
}

//---------------------------------------------------------------------------
//	Member Function of Rest Class
//---------------------------------------------------------------------------	
Rest::Rest()
{}

//---------------------------------------------------------------------------
//	Member Function of Rest Class
//---------------------------------------------------------------------------	
Rest::Rest( string *pstr_newname)
{
	name = *pstr_newname;
}

//---------------------------------------------------------------------------
//	Member Function of Rest Class
//---------------------------------------------------------------------------	
bool Rest::growTree()
{
	delaytime = atoi( name.c_str());
	return true;
}

//---------------------------------------------------------------------------
//	Member Function of Rest Class
//---------------------------------------------------------------------------	
void Rest::play()
{
	int counter = 0;
	int loop	= delaytime / 500;
	int	remain	= delaytime % 500;

	while( counter < loop)
	{
		if( stopAllThread)	return;
		SDL_Delay( 500);
		counter++;
	}	
	SDL_Delay( remain);
}

//---------------------------------------------------------------------------
//	Member Function of Rest Class
//---------------------------------------------------------------------------
void Rest::setPosition4Note( Sint16 new_angle, Uint8 new_dist){}
Rest::~Rest()
{
}

//---------------------------------------------------------------------------
//	Member Function of RandomRest Class
//---------------------------------------------------------------------------
RandomRest::RandomRest()
{
	rand_max = 0;
	rand_min = 0;
}

//---------------------------------------------------------------------------
//	Member Function of RandomRest Class
//---------------------------------------------------------------------------
RandomRest::RandomRest( string *pstr_newname)
{
	name = *pstr_newname;
}

//---------------------------------------------------------------------------
//	Member Function of RandomRest Class
//---------------------------------------------------------------------------
void RandomRest::play()
{
	int delay = rand.GetRandom( rand_min, rand_max);

	Rest::delaytime = delay;
	if( !stopAllThread)	Rest::play();
}


//---------------------------------------------------------------------------
//	Member Function of RandomRest Class
//---------------------------------------------------------------------------
void RandomRest::setPosition4Note( Sint16 new_angle, Uint8 new_dist)
{//	do nothing
}

//---------------------------------------------------------------------------
//	Member Function of RandomRest Class
//---------------------------------------------------------------------------
bool RandomRest::growTree()
{
	list<string> lst_str;
	list<string>::iterator it;

	int temp = 0;

	devideStringByStringEx( &name, "to", &lst_str, false);
	it			= lst_str.begin();
	rand_min	= atoi( (*it).c_str());
	it++;
	rand_max	= atoi( (*it).c_str());

	if( rand_min > rand_max)
	{
		temp		= rand_max;
		rand_max	= rand_min;
		rand_min	= temp;
	}
	return true;
}

//---------------------------------------------------------------------------
//	Member Function of RandomRest Class
//---------------------------------------------------------------------------
RandomRest::~RandomRest()
{
}

//---------------------------------------------------------------------------
//	Member Function of Note Class
//---------------------------------------------------------------------------
Note::Note()
{
}

//---------------------------------------------------------------------------
//	Member Function of Note Class
//---------------------------------------------------------------------------

Note::Note( string *pstr_newname)
{
	name = *pstr_newname;
}

//---------------------------------------------------------------------------
//	Member Function of Note Class
//---------------------------------------------------------------------------

void Note::play()
{
	if( !stopAllThread)		sound.play();
}

//---------------------------------------------------------------------------
//	Member Function of Note Class
//---------------------------------------------------------------------------
void Note::setPosition4Note( Sint16 new_angle, Uint8 new_dist)
{
	sound.setPosition( new_angle, new_dist);
}

//---------------------------------------------------------------------------
//	Member Function of Note Class
//---------------------------------------------------------------------------
bool Note::growTree()
{
	string wavefile = name.substr( 1, name.length() - 2);

	if( !sound.load( &wavefile))
	{
		cout << "Error : Note::sound.load() returns false" << endl;
		return false;
	}

	return true;
}

//---------------------------------------------------------------------------
//	Member Function of Note Class
//---------------------------------------------------------------------------
Note::~Note()
{
}

//---------------------------------------------------------------------------
//	Member Function of MultiPattern Class
//---------------------------------------------------------------------------
MultiPattern::MultiPattern()
{
}

//---------------------------------------------------------------------------
//	Member Function of MultiPattern Class
//---------------------------------------------------------------------------
MultiPattern::MultiPattern( string *pstr_newname)
{
	name = *pstr_newname;
}

//---------------------------------------------------------------------------
//	Member Function of MultiPattern Class
//	set parameters from string seaquence.
//	The string seaquence is stored as a member name
//	for ex. (a:b=1:2),c
//	=>	MultiaPattern( (a:b=1:2))
//		  + MultiPattern( c)
//---------------------------------------------------------------------------
bool MultiPattern::growTree()
{
	list<string> lst_str;
	list<string>::iterator it;
	string str_sample;
	// Never fail cause already tested by diagnoseType()
	devideStringByStringEx( &name, ",", &lst_str, true);
	it = lst_str.begin();
	while( it != lst_str.end())
	{
		str_sample = *it++;
		lst_ppattern.push_back( new Pattern( &str_sample));
	}
	it_ppattern = lst_ppattern.begin();
	while( it_ppattern != lst_ppattern.end())
	{
		if( !(*it_ppattern)->growTree())
		{
			cout << "MultiPattern::growTree Error" << endl;
			return false;
		}
		it_ppattern++;
	}
	return true;
}

//---------------------------------------------------------------------------
//	Member Function of MultiPattern Class
//---------------------------------------------------------------------------	

void MultiPattern::play()
{
	it_ppattern = lst_ppattern.begin();
	while( it_ppattern != lst_ppattern.end() && !stopAllThread)
	{
		(*it_ppattern)->play();
		it_ppattern++;
	}
}

//---------------------------------------------------------------------------
//	Member Function of MultiPattern Class
//---------------------------------------------------------------------------	
void MultiPattern::setPosition4Note( Sint16 new_angle, Uint8 new_dist)
{
	it_ppattern = lst_ppattern.begin();
	while( it_ppattern != lst_ppattern.end())
	{
		(*it_ppattern)->setPosition4Note( new_angle, new_dist);
		it_ppattern++;
	}
}

//---------------------------------------------------------------------------
//	Member Function of MultiPattern Class
//---------------------------------------------------------------------------
MultiPattern::~MultiPattern()
{

	it_ppattern = lst_ppattern.begin();
	while( it_ppattern != lst_ppattern.end())
	{
    	if( *it_ppattern != NULL)	(*it_ppattern)->~Pattern();
		it_ppattern++;
	}
}

//---------------------------------------------------------------------------

//	Member Function of ProbabilityPattern Class
//---------------------------------------------------------------------------
ProbabilityPattern::ProbabilityPattern()
{
	sum = 0;
}

//---------------------------------------------------------------------------
//	Member Function of ProbabilityPattern Class
//---------------------------------------------------------------------------
ProbabilityPattern::ProbabilityPattern( string *pstr_newname)


{
	name	= *pstr_newname;
	sum		= 0;
}
//---------------------------------------------------------------------------
//	Member Function of ProbabilityPattern Class
//---------------------------------------------------------------------------
bool ProbabilityPattern::growTree()
{
	list<string>			lst_str;
	list<string>::iterator	it_str, it_obj, it_rate;

	list<string> lst_str_prob_obj;
	list<string> lst_str_prob_rate;

	int converted		= 0;
	string str_obj_n	= name;	
	//-------------------------------------------------------------------------
	//	Devide string by '=" into left hand and right hand.
	//	Never failt becouse isProbabilityPattern() has already checked.
	//-------------------------------------------------------------------------
	devideStringByStringEx( &str_obj_n, "=", &lst_str, true);

	//-------------------------------------------------------------------------
	//	Devide left hand by ':'. Never fail w/ same reason
	//-------------------------------------------------------------------------
	it_str		= lst_str.begin();
	str_obj_n	= *it_str++;
	devideStringByStringEx( &str_obj_n, ":", &lst_str_prob_obj, true);
	//-------------------------------------------------------------------------
	//	Devide right hand by ':'. Never fail too.
	//-------------------------------------------------------------------------
	str_obj_n	= *it_str;
	devideStringByStringEx( &str_obj_n, ":", &lst_str_prob_rate, true);

	//-------------------------------------------------------------------------
	//	Initialize these iterator for later access
	//-------------------------------------------------------------------------
	it_obj		= lst_str_prob_obj.begin();
	it_rate		= lst_str_prob_rate.begin();

	//-------------------------------------------------------------------------
	//	Store the probability into the list.
	//	Possible to fail due to isProbabilityPattern() doesn't check this part.
	//-------------------------------------------------------------------------
	while( it_rate != lst_str_prob_rate.end())
	{
		converted	= atoi( (*it_rate).c_str());
		if( converted == -1)	return false;
		sum	+= converted;
		lst_probability.push_back( converted);
		it_rate++;
	}
	//-------------------------------------------------------------------------
	//	Create Pattern objects
	//-------------------------------------------------------------------------
	it_obj	= lst_str_prob_obj.begin();
	it_rate	= lst_str_prob_rate.begin();
	while( it_obj != lst_str_prob_obj.end())
	{
		str_obj_n = *it_obj++;
		lst_ppattern.push_back( new Pattern( &str_obj_n));
	}
	it_ppattern = lst_ppattern.begin();
	while( it_ppattern != lst_ppattern.end())
	{
		if( !(*it_ppattern)->growTree())
			return false;
		it_ppattern++;
	}
  return true;	
}

//---------------------------------------------------------------------------
//	Member Function of ProbabilityPattern Class
//---------------------------------------------------------------------------
void ProbabilityPattern::play()
{
	it_ppattern		= lst_ppattern.begin();
	it_probability	= lst_probability.begin();
	int randnum		= rand.GetRandom( 0, sum - 1);
	int accumulate	= 0;

	while( it_ppattern != lst_ppattern.end() && !stopAllThread)
	{
		accumulate += *it_probability;
		if( randnum < accumulate)
		{
			(*it_ppattern)->play();
			break;
		}

		it_probability++;
		it_ppattern++;
	}
}

//---------------------------------------------------------------------------
//	Member Function of ProbabilityPattern Class
//---------------------------------------------------------------------------
ProbabilityPattern::~ProbabilityPattern()
{
}

//---------------------------------------------------------------------------
//	Constructor of Alies Class with Parameters
//---------------------------------------------------------------------------
Alies::Alies()
{
	ppattern = NULL;
}
//---------------------------------------------------------------------------
//	Constructor of Alies Class with Parameters
//---------------------------------------------------------------------------
Alies::Alies( string *pstrname, Pattern *ppatnew)
{
	name 		= *pstrname;
	ppattern	= ppatnew;
}
Alies::~Alies()
{
	if( ppattern != NULL)	ppattern->~Pattern();
}

//---------------------------------------------------------------------------
//	Member Function of Song Class
//---------------------------------------------------------------------------
Song::Song(){

	ppattern	= NULL;
	timesplay	= 1;
}
//---------------------------------------------------------------------------
//	Member Function of Song Class
//---------------------------------------------------------------------------
Song::Song( string *pstr, Pattern *pnewpat, unsigned int new_timesplay){
	name		= *pstr;
	ppattern	= pnewpat;
	timesplay	= new_timesplay;
}

//---------------------------------------------------------------------------
//	Member Function of Song Class
//	Here, we don't have to delete ppattern. Because super class of this class
//	Song takes care of that in it's destructor.
//---------------------------------------------------------------------------
Song::~Song(){
	
	it_alies = lst_alies.begin();
	while( it_alies != lst_alies.end())
	{
		(*it_alies)->~Alies();
		it_alies++;
	}
}
//---------------------------------------------------------------------------
//	Member Function of Song Class
//---------------------------------------------------------------------------
void Song::addSubpatterns( string *str_name, Pattern *ppat_newsub)
{
	//cout << "Alies " << *str_name << " = " << ppat_newsub->getName() << " is stored in list now size is "
	//	<< lst_alies.size() << endl;
	lst_alies.push_back( new Alies( str_name, ppat_newsub));
}

//---------------------------------------------------------------------------
//	Song::growTree()
//	Create a tree. First grow sub trees then main tree, ppattern.
//	Subtrees are listed in a list lst_ppat_alies.
//	If this fail with some reason such like syntax mistake in string in
//	Pattern Object and can't be able to grow, this returns fail.
//	Pattern objects in ppattern and lst_ppat_alies are created by "new" so
//	those are destructed immediately by calling ~Song()
//---------------------------------------------------------------------------
bool Song::growTree()
{

	if( lst_alies.size() == 0)	return false;

	lst_table_alies = this->lst_alies;
	it_table_alies	= this->it_alies;

	it_alies = this->lst_alies.begin();
	while( this->it_alies != this->lst_alies.end())
	{
		if( !(*(this->it_alies))->ppattern->growTree())	return false;
		(this->it_alies)++;
	}
	if( !ppattern->growTree())			return false;
	
	return true;
}
//---------------------------------------------------------------------------
//	( (a,b),c:d=1:2)) => MultiPattern
//---------------------------------------------------------------------------
bool isMultiPattern( string *pstr)
{
	list<string> lst_str;
	string sample = *pstr;

	if( !devideStringByStringEx( &sample, ",", &lst_str, true))
		return false;
	if( lst_str.size() <= 1)
		return false;
  return true;
}

//---------------------------------------------------------------------------
//	Check if the string represents ProbabilityPattern or not
//---------------------------------------------------------------------------
bool isProbabilityPattern( string *pstr)// this comes next to MultiPattern
{
	list<string> 			lst_str,
							lst_str_obj,
							lst_str_rate;
	list<string>::iterator	it;
	string					sample = *pstr;

	if( !devideStringByStringEx( &sample, "=", &lst_str, true))
		return false;

	if( lst_str.size() != 2)
		return false;
	it		= lst_str.begin();
	sample	= *it;
	if( !devideStringByStringEx( &sample, ":", &lst_str_obj, true))
		return false;
	it++;
	sample	= *it;
	if( !devideStringByStringEx( &sample, ":", &lst_str_rate, true))
		return false;
	return true;
}
//---------------------------------------------------------------------------
//	Check if the string represents Note or not
//---------------------------------------------------------------------------

bool isNote( string *pstr)
{
	if( countString( pstr, "\"") == 2)
		return true;
	return false;
}

//---------------------------------------------------------------------------
//	Check if the string represents RandomRest or not
//---------------------------------------------------------------------------
bool isRandomRest( string *pstr)// this comes
{
	list<string>			lst_str;
	list<string>::iterator	it;
    string					str_num;

	if( !devideStringByStringEx( pstr, "to", &lst_str, false))
		return false;
	if( lst_str.size() != 2)	return false;
	it = lst_str.begin();

	str_num = *it++;
	if( !isNumber( &str_num))	return false;
	str_num = *it;
	if(	!isNumber( &str_num))	return false;
	return true;
}

//---------------------------------------------------------------------------
//	Diagnose what the string seaquence represents.
//	input string shouldn't surrounded by parenthesis
//---------------------------------------------------------------------------
TYPE_DIAGNOSE_TYPE diagnoseType( string *pstr)
{
	// When devideStringByStringEx returns only 1 string => undevidable
	// => Pattern
	if( isMultiPattern( pstr))
	{
		return TYPE_DIAGNOSE_MULTIPATTERN;
	}
	// Contains ':' ->ProbabilityNote
	else if( isProbabilityPattern( pstr))
	{
		return TYPE_DIAGNOSE_PROBABILITYPATTERN;
	}		
	// 1 word && no '"' && number && non-variable -> RandomRest
	else if( isRandomRest( pstr))
	{
		return TYPE_DIAGNOSE_RANDOMREST;
	}
	// Surrounded by "," -> Note
	else if( isNote( pstr))
	{
		return TYPE_DIAGNOSE_NOTE;
	}
	else if( isNumber( pstr))
	{
		return TYPE_DIAGNOSE_REST;
	}
	else
	{
		return TYPE_DIAGNOSE_LABEL;
	}
	return 999;
}

//---------------------------------------------------------------------------
//	Refere table and find pattern object as name matched.
//---------------------------------------------------------------------------
Pattern* searchSongList( string *pstr_song_name)
{
	it_table_alies	= lst_table_alies.begin();
	while( it_table_alies != lst_table_alies.end())
	{
		if( pstr_song_name->compare( (*it_table_alies)->name) == 0)
			return (*it_table_alies)->ppattern;
		it_table_alies++;
	}
	return NULL;
}

//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
/*void testDiagnoseType()
{
	//string line = "a:b:c:d=1:2:3:4";	// ProbabilityPattern
	//string line = " \"a.wav\", \"b.wav\", \"c.wav\"";// MultiPattern
	//string line = "100to200";// RandomRest
	//string line = "\"hoge hoge.wav\"";
	//string line = "(a:b=1:2),\"ho ge.wav\",AAA";
	string line = "AAA";
	diagnoseType( &line);
}*/


