/***************************************************************************
                          pattern.h  -  description
                             -------------------
    begin                : Wed Jun 26 2002
    copyright            : (C) 2002 by Nobuhiro Kuroiwa
    email                : nobu@iastate.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef PATTERN_H
#define PATTERN_H
/**
  *@author Nobuhiro Kuroiwa
  */

#define TYPE_DIAGNOSE_TYPE					unsigned int
#define TYPE_DIAGNOSE_PATTERN 				0
#define TYPE_DIAGNOSE_NOTE					1
#define	TYPE_DIAGNOSE_REST					2
#define	TYPE_DIAGNOSE_RANDOMREST			3
#define	TYPE_DIAGNOSE_MULTIPATTERN			4
#define	TYPE_DIAGNOSE_PROBABILITYPATTERN	5
#define TYPE_DIAGNOSE_LABEL					6


#include <list>
#include <string>
#include "smplsnd.h"
#include "pattern.h"
#include "crand.h"

using namespace std;

class Pattern {
public:
	Pattern();
	Pattern( string *pstr_newname);
	virtual	~Pattern();
	virtual	bool growTree();
	virtual	void play();
	string	getName() const;
	virtual	void setPosition4Note( Sint16 new_angle, Uint8 new_dist);
protected:
	Pattern *ppattern;
	string	name;
	bool	own_ppat;
};


class Note : public Pattern
{
public:
	Note();
	Note( string *pstr_newname);
	bool growTree();
	void play();
	void setPosition4Note( Sint16 new_angle, Uint8 new_dist);
	~Note();
private:
	SimpleSound	sound;
};

class Rest : public Pattern
{
public:
	Rest();
	Rest(	string *pstr_newname);
	bool	growTree();
	void	play();
	void	setPosition4Note( Sint16 new_angle, Uint8 new_dist);
	~Rest();
protected:
	unsigned int delaytime;
};

class RandomRest : public Rest// should be changed
{
public:
	RandomRest();
	RandomRest( string *pstr_newname);
	bool growTree();
	void play();
	void setPosition4Note( Sint16 new_angle, Uint8 new_dist);
	~RandomRest();
private:
	int		rand_min, rand_max;
	CRand	rand;
};


class MultiPattern : public Pattern
{
public:
	MultiPattern();
	MultiPattern(	string *pstr_newname);
	void play();
	bool growTree();
	void setPosition4Note( Sint16 new_angle, Uint8 new_dist);
	~MultiPattern();
protected:
	std::list<Pattern*>				lst_ppattern;
	std::list<Pattern*>::iterator	it_ppattern;
};

class ProbabilityPattern : public MultiPattern
{
public:
	ProbabilityPattern();
	ProbabilityPattern( string *pstr_newname);
	void play();
	bool growTree();
	~ProbabilityPattern();
private:
	std::list<int>				lst_probability;
	std::list<int>::iterator	it_probability;
	int		sum;
	CRand	rand;
};

class Alies{
public:
	Pattern		*ppattern;
	string		name;
	Alies();
	Alies( string *pstrname, Pattern *ppatnew);
	~Alies();
};


class Song : public Alies{
public:
	Song();
	Song( string *pstr, Pattern *pnewpat, unsigned int new_timesplay = 1);
	void	addSubpatterns( string *str_name, Pattern *ppat_newsub);
	bool	growTree();
	~Song();
	unsigned int				timesplay;
	string						name;
	std::list<Alies *> 			lst_alies;
	std::list<Alies*>::iterator	it_alies;

};


TYPE_DIAGNOSE_TYPE
			diagnoseType( string *pstr);
bool		isMultiPattern( string *pstr);
bool		isNote( string *pstr);
bool		isRest( string *pstr);
bool		isProbabilityPattern( string *pstr);
Pattern*	searchSongList( string *pstr_song_name);

//void	testDiagnoseType();// test function. can be removed
#endif


