/***************************************************************************
                          suzumusi.h  -  description
                             -------------------
    begin                : Mon Jul 1 2002
    copyright            : (C) 2002 by Nobuhiro Kuroiwa
    email                : nobu@iastate.edu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef	SUZUMUSI_H
#define	SUZUMUSI_H

using namespace std;

#include <iostream>
#include <string>
#include <fstream>
#include <list>
#include "SDL/SDL.h"
#include "SDL/SDL_thread.h"
#include "pattern.h"
#include "stranlz.h"
#include "suzumusi.h"

bool	readSoundParameters(
							string			*pstr,
							unsigned int	*ptimes,
							Sint16			*pangle,
							Uint8			*pdist
);
bool	analyzeFileString(		string			*pstr,
							list<Song*>		*plst_psongs
);
Song	*createSong( 
							string			*pstr_songname,
							string			*pstr,
							unsigned int	*ptimesplay,
							Sint16			*pangle,
							Uint8			*pdistance
);
void		cleanUp();
void		letsDance();
void		stopDance();
int			threadPlay( void *psong2play);

list<Song*>							lst_psongs;// playlist
list<SDL_Thread*> 					lst_thread;
list<SDL_Thread*>::iterator		 	it_thread;
list<Song*>::iterator				it_psongs;

bool		stopAllThread	= false;
SDL_sem*	my_sem			= NULL;

//---------------------------------------------------------------------------
//	In the main function, when this initSuzumusi returns false,
//	cleanUp() is called in main. So we just return false.
//---------------------------------------------------------------------------
bool initSuzumusi( char *pchar)
{
	string str_file;
	//-----------------------------------------------------------------------
	//	Load text file to string
	//-----------------------------------------------------------------------
	if(	!loadFile( pchar, &str_file))
	{
	  cout << "Unable to load file " << pchar << endl;
		return false;
	}
	//-----------------------------------------------------------------------
	//	Erase Space, Erase Comment and Erase \n
	//-----------------------------------------------------------------------
	eraseComment( &str_file);
	eraseSpace( &str_file);
	eraseString( &str_file, "\n");
	//-----------------------------------------------------------------------
	//	SDL Initialization Stuff
	//-----------------------------------------------------------------------
	if( SDL_Init( SDL_INIT_AUDIO) == -1)
	{
		cout << "SDL initialization error : " << SDL_GetError() << endl;
		return false;
	}

	if( Mix_OpenAudio(
				AUDIO_FREQUENCY,
				AUDIO_FORMAT,
				AUDIO_MODE,
				AUDIO_BUFFERSIZE)
			 == -1
	)
	{
 		cout << "Unable to open audio! : " << Mix_GetError() << endl;
 		return false;
	}
	//-----------------------------------------------------------------------
	//	16 is not enough for some heavy user maybe...
	//	This value should be from play list. I'll write those code someday...
	//-----------------------------------------------------------------------
	Mix_AllocateChannels( 16);
	//-----------------------------------------------------------------------
	//	Analyze the strings and make Song objects which has tree for each.
	//-----------------------------------------------------------------------
	if( !analyzeFileString( &str_file, &lst_psongs))
	{
		cout << "Analyzing File String Failed" << endl;
		return false;
	}
	//	Create a Semaphore
	my_sem = SDL_CreateSemaphore( 1);
	if( my_sem == NULL)	cout << "Failed to Create a Semaphore" << endl;
	return true;
}

//---------------------------------------------------------------------------
//	readSoundParameters takes care of the sound parameter parts of playlist
//	such like _param = ( _infinite, 90, 200);
//	The 1st word of parenthesis means number of looping of the song.
//	The 2nd word of parenthesis means angle of sound source. Front is 0. Back
//	is 180. Right is 90. Left is 270.
//	The 3rd word of parenthesis means distance between sound source and
//	listner(us).
//---------------------------------------------------------------------------
bool readSoundParameters(
	string *pstr, unsigned int* ptimes, Sint16 *pangle, Uint8 *pdist
)
{
	list<string>			lst_str;
	list<string>::iterator	it;

	int		converted;
	bool	ret = true;

	pealParenthesis( pstr);

	devideStringByStringEx( pstr, ",", &lst_str, false);	
	//-----------------------------------------------------------------------
	//	parameter consists of 3 words, times to play, angle and distance
	//-----------------------------------------------------------------------
	if( lst_str.size() < 3)
	{
		cout << "too few arguments for param" << endl;
		cout << "playlist.txt is set to parameter" << endl;
		return false;
	}

	//-----------------------------------------------------------------------
	//	If the 1st argument of parameter which represents times to loop
	//	is _infinite, the 1000 is set and will be deal as infinite times
	//	in threadPlay()
	//-----------------------------------------------------------------------
	it = lst_str.begin();
	converted = atoi( (*it).c_str());
	if( *it == "_infinite")
	{
		//	timesloop is set to infinity. 1000 is a label as infinity.
		converted = 1000;
	}
 	else if( converted < 0 && 999 <= converted)// reserve 1000 for infinity
	{
		cout	<<	"can't understand as number of times loop \""
				<< converted << "\"" <<	endl
				<< "number of times loop set to 1"	<< endl;
		//	times loop is set to 1 if the new value is quite out of our range.
		converted	= 1;
		ret			= false;
	}

	*ptimes = converted;

	//-----------------------------------------------------------------------
	//	Convert second word to integer. It should between 0 to 360.
	//-----------------------------------------------------------------------
	it++;
	converted = atoi( (*it).c_str());
	if( converted < 0 || 360 < converted)
	{
		cout << "can't understand as angle \"" << converted << "\"" << endl;
		converted	= AUDIO_DEFAULT_ANGLE;
		ret			= false;
	}
	//	Set angle to new value
	*pangle = converted;

	//-----------------------------------------------------------------------
	//	Convert third word to integer. It should between 0 to 255.
	//-----------------------------------------------------------------------
	it++;
	converted = atoi( (*it).c_str());

	if( converted < 0 || 255 < converted)
	{
		cout	<< "can't understand as distance " << converted
				<< ", must be 0-255." << endl;
		converted	= AUDIO_DEFAULT_DISTANCE;
		ret			= false;
	}
	//	Set distance to new value
	*pdist = converted;

	return ret;
}

typedef std::list<string> StrLst;
//---------------------------------------------------------------------------
//	String must be like this "aaa = \"a.wav\"; _main = (aaa, \"b.wav\")"
//	combinePattern() completes tree of one song.
//	This function create a song from string seaquence
//	pstr is data of a song like "suzumusi1 = ( ... _main(...))"
//---------------------------------------------------------------------------
Song *createSong( 
					string			*pstr_songname,
					string			*pstr,
					unsigned int	*ptimesplay,
					Sint16			*pangle,
					Uint8			*pdistance
)
{
	list<StrLst>			lst_lststr;
	list<StrLst>::iterator	it_lststr;
	list<string>			lst_str;
	list<string>::iterator	it_str;
	string					str_name;
	string					str_data;
	Song					*psong_ret = new Song();

	pealParenthesis( pstr);

	if( !devideDevided( pstr, ";", "=", &lst_lststr))
	{
		cout	<< "devideDevide() failed : " << *pstr << " \";\" and "
				<< "\"=\""	<< endl;
		return NULL;
	}
	
	it_lststr = lst_lststr.begin();

	while( it_lststr != lst_lststr.end())
	{
		lst_str	= *it_lststr++;
		it_str	= lst_str.begin();

    	if( lst_str.size() != 2)
		{
			cout	<< "Error in combinePattern() : lst_str.size() != 2"
					<< endl;
			while( it_str != lst_str.end())
			{
				cout << "\tlst_str is : \"" << *it_str << "\"" << endl;
				it_str++;
			}
			return NULL;
		}

		while( it_str != lst_str.end())
		{
			str_name = *it_str++;
			str_data = *it_str++;
			// main pattern
			if( str_name == "_main")
			{
				psong_ret->name = *pstr_songname;	// must be "_main"
				// set new Pattern object to psong but not yet growTreed.
				psong_ret->ppattern = new Pattern( &str_data);
			}
			// parameters
			else if( str_name == "_param")
			{
				// implement code for parametr here, set parameters from
				if( !readSoundParameters(
											&str_data,
											ptimesplay,
											pangle,
											pdistance
										)
				)
					cout << "parameters set to default" << endl;
			}
			// alias : sub pattern
			else
			{
				psong_ret->addSubpatterns(
											&str_name,
											new Pattern( &str_data)
										);
			}
		}
	}
	 return psong_ret;
	// Need to call growTree() for newly created Pattern object
}
//---------------------------------------------------------------------------
//	analyzeFileString() digests string of playlist and create number of
//	Song objects as needed. The Song objects are stored in table
//	list<Song>* plst_psongs
//---------------------------------------------------------------------------
bool analyzeFileString( string *pstr, list<Song*> *plst_psongs)
{
	list<StrLst>			lst_lststr;
	list<string>			lst_str;
	list<StrLst>::iterator	it_lststr;
	list<string>::iterator	it_str;
	string					str_asong;
	string					songname;
	unsigned int			timesplay	= 1;
	Sint16					angle		= AUDIO_DEFAULT_ANGLE;
	Uint8					distance	= AUDIO_DEFAULT_DISTANCE;
	Song *psong;

	// devide "suzumusi=(...);matumusi=(...);" into
	// "suzumusi=(...)" and "matumusi=(...)"
	if( !devideDevided( pstr, ";", "=", &lst_lststr))
	{
		cout << "Unable to devide \"" << *pstr << "\" by \"=\" then by \";\""
			 << "\nString must be this form n1:n2:...=x1:x2:..." << endl;
	}
	it_lststr = lst_lststr.begin();

	while( it_lststr != lst_lststr.end())
	{
		lst_str		= *it_lststr++;
		it_str		= lst_str.begin();
		songname	= *it_str++;
		str_asong	= *it_str;
		psong = createSong(
							&songname,
							&str_asong,
							&timesplay,
							&angle,
							&distance
						);
		if( psong == NULL)
		{
			cout << "can't create Song " << songname << endl;
			return false;
		}
		if( !psong->growTree())
		{
			cout << songname << " : growTree error" << endl;
			psong->~Song();
			return false;
		}
		//	Add newly created Pattern Object ppat_new is stored in a Table
		//  plst_psongs.
		psong->ppattern->setPosition4Note( angle, distance);
		psong->timesplay = timesplay;
		plst_psongs->push_back( psong);
	}

	return true;
}

//---------------------------------------------------------------------------
//	cleanUp() releases all resources we reserved or allocated.
//	This must be called whenever before this program terminates.
//---------------------------------------------------------------------------
void cleanUp()
{
	//-----------------------------------------------------------------------
	//	Stop all threads
	//-----------------------------------------------------------------------
	stopDance();
	//-----------------------------------------------------------------------
	//	Halt All Channels
	//-----------------------------------------------------------------------
	cout << "halting " << Mix_Playing( -1) << " channels...";
	if( Mix_Playing( -1) > 0)
	{
	  	Mix_HaltChannel( -1);
	}
	Mix_AllocateChannels( 0);
	cout << "done." << endl;
	//-----------------------------------------------------------------------
	//	Free Memory
	//-----------------------------------------------------------------------
	cout << "cleaning memory...";
	//-----------------------------------------------------------------------
	//	Free Songs
	//-----------------------------------------------------------------------
	it_psongs = lst_psongs.begin();
	while( it_psongs != lst_psongs.end())
	{
		(*it_psongs)->~Song();
		it_psongs++;
		break;
	}
	cout << "done" << endl;
	//-----------------------------------------------------------------------
	//	Close Audio
	//-----------------------------------------------------------------------
	Mix_CloseAudio();
	cout << "audio closed" << endl;
	SDL_Quit();
	cout << "suzumusi is terminated" << endl;
}

//---------------------------------------------------------------------------
//	threadPlay() is lounched as a thread. There are multiple number of this
//	thread will be lounched. The number is as many as number of songs in
//	play list. So this cout statement can be interrupted by another threads.
//---------------------------------------------------------------------------
int threadPlay( void *psong2play)
{
	Song			*psong	= (Song*)psong2play;
	unsigned int	counter	= 0;

	//-----------------------------------------------------------------------
	//	Display how many times to loop. If it's 1000, display infinite.
	//-----------------------------------------------------------------------
    SDL_SemWait( my_sem);
	cout << "playing ";
	if( psong->timesplay == 1000)	cout << "infinite";
	else							cout << psong->timesplay;
	cout << " time(s) of \"" << psong->name << "\"" << endl;
    SDL_SemPost( my_sem);
	//-----------------------------------------------------------------------
	//	Loop
	//-----------------------------------------------------------------------
	while( counter < psong->timesplay)
	{
		if( stopAllThread)	break;
		psong->ppattern->play();
		//-------------------------------------------------------------------
		//	If timesplay is 1000, counter doesn't increas meaning that the
		//	loop runs forever untill stopAllThread is true.
		//-------------------------------------------------------------------
		if( psong->timesplay != 1000)	counter++;
	}
    SDL_SemWait( my_sem);
	cout << psong->name << " done" << endl;
    SDL_SemPost( my_sem);
	return 0;
}

//---------------------------------------------------------------------------
//	letsDance() lounches threadPlay() as a thread. The threads are created
//	as many as number of Songs in play list. One song is assigned in one
//	thread.
//---------------------------------------------------------------------------
void letsDance()
{
	it_psongs = lst_psongs.begin();

	//-----------------------------------------------------------------------
	//	Create a thread for one Song one thread.
	//-----------------------------------------------------------------------
	while( it_psongs != lst_psongs.end())
	{
		lst_thread.push_back( SDL_CreateThread( threadPlay,(void*)*it_psongs));
		it_psongs++;
	}
}

//---------------------------------------------------------------------------
//	stopDance() set the flag stopAllThread to true. Threads are checking
//	this flag and finish it self in a moment. So stopDance() waits for the
//	threads finished.
//---------------------------------------------------------------------------
void stopDance()
{

	cout << "terminating all threads...\n";
	stopAllThread	= true;
	it_thread		= lst_thread.begin();

	while( it_thread != lst_thread.end())
	{
		SDL_WaitThread( *it_thread, NULL);
		it_thread++;
	}
}

#endif

